@@ -4,7 +4,7 @@
 
 ;; Author: "Francis J. Wright" <F.J.Wright@qmul.ac.uk>
 ;; Maintainer: emacs-devel@gnu.org
-;; Version: 1.12
+;; Version: 1.18
 ;; Package-Requires: ((emacs "24.1") (cl-lib "0.5"))
 ;; Keywords: convenience
 
@@ -146,7 +146,10 @@ Set by customizing `csv-separators' -- do not set directly!")
 For example: (\",\"), the default, or (\",\" \";\" \":\").
 Neighbouring fields may be separated by any one of these characters.
 The first is used when inserting a field separator into the buffer.
-All must be different from the field quote characters, `csv-field-quotes'."
+All must be different from the field quote characters, `csv-field-quotes'.
+
+Changing this variable with `setq' won't affect the current Emacs
+session.  Use `customize-set-variable' instead if that is required."
   ;; Suggested by Eckhard Neber <neber@mwt.e-technik.uni-ulm.de>
   :type '(repeat string)
   ;; FIXME: Character would be better, but in Emacs 21.3 does not display
@@ -235,7 +238,8 @@ Auto-alignment means left align text and right align numbers."
 
 (defcustom csv-align-padding 1
   "Aligned field spacing: must be a positive integer.
-Number of spaces used by `csv-align-mode' and `csv-align-fields' after separators."
+Number of spaces used by `csv-align-mode' and `csv-align-fields'
+after separators."
   :type 'integer)
 
 (defcustom csv-header-lines 0
@@ -662,8 +666,22 @@ point or marker arguments, BEG and END, delimiting the region."
 (defun csv-end-of-field ()
   "Skip forward over one field."
   (skip-chars-forward " ")
-  (if (eq (char-syntax (following-char)) ?\")
-      (goto-char (scan-sexps (point) 1)))
+  ;; If the first character is a double quote, then we have a quoted
+  ;; value.
+  (when (eq (char-syntax (following-char)) ?\")
+    (forward-char)
+    (let ((ended nil))
+      (while (not ended)
+	(cond ((not (eq (char-syntax (following-char)) ?\"))
+	       (forward-char 1))
+	      ;; According to RFC-4180 (sec 2.7), quotes inside quoted strings
+	      ;; are quoted by doubling the quote char: a,"b""c,",d
+	      ;; FIXME: Maybe we should handle this via syntax-propertize?
+              ((let ((c (char-after (1+ (point)))))
+                 (and c (eq (char-syntax c) ?\")))
+	       (forward-char 2))
+	      (t
+	       (setq ended t))))))
   (skip-chars-forward csv--skip-chars))
 
 (defun csv-beginning-of-field ()
@@ -818,11 +836,15 @@ the mode line after `csv-field-index-delay' seconds of Emacs idle time."
 
 (defun csv--field-index ()
   (save-excursion
-    (let ((lbp (line-beginning-position)) (field 1))
-      (while (re-search-backward csv-separator-regexp lbp 'move)
-	;; Move as far as possible, i.e. to beginning of line.
+    (let ((start (point))
+	  (field 0))
+      (beginning-of-line)
+      (while (< (point) start)
+	(csv-end-of-field)
+	(unless (eolp)
+	  (forward-char 1))
 	(setq field (1+ field)))
-      (unless (csv-not-looking-at-record) field))))
+      field)))
 
 (defun csv-field-index ()
   "Construct `csv-field-index-string' to display in mode line.
@@ -1248,9 +1270,7 @@ When called non-interactively, BEG and END specify region to process."
 	      (forward-line)
 	    (let ((lep (line-end-position)))
 	      (push
-	       (split-string
-		(buffer-substring-no-properties (point) lep)
-		csv-separator-regexp)
+	       (csv--collect-fields lep)
 	       rows)
 	      (delete-region (point) lep)
 	      (or (eobp) (delete-char 1)))))
@@ -1289,6 +1309,26 @@ When called non-interactively, BEG and END specify region to process."
 	;; Re-do soft alignment if necessary:
 	(if align (csv-align-fields nil (point-min) (point-max)))))))
 
+(defun csv--collect-fields (row-end-position)
+  "Collect the fields of a row.
+Splits a row into fields, honoring quoted fields, and returns
+the list of fields.  ROW-END-POSITION is the end-of-line position.
+point is assumed to be at the beginning of the line."
+  (let ((csv-field-quotes-regexp (apply #'concat `("[" ,@csv-field-quotes "]")))
+	(row-text (buffer-substring-no-properties (point) row-end-position))
+	fields field-start)
+    (if (not (string-match csv-field-quotes-regexp row-text))
+	(split-string row-text csv-separator-regexp)
+      (save-excursion
+	(while (< (setq field-start (point)) row-end-position)
+	  (csv-forward-field 1)
+	  (push
+	   (buffer-substring-no-properties field-start (point))
+	   fields)
+	  (if (memq (following-char) csv-separator-chars)
+	      (forward-char)))
+	(nreverse fields)))))
+
 (defvar-local csv--header-line nil)
 (defvar-local csv--header-hscroll nil)
 (defvar-local csv--header-string nil)
@@ -1336,7 +1376,8 @@ If there is already a header line, then unset the header line."
       (jit-lock-fontify-now (point) (line-end-position))
       ;; Not sure why it is sometimes nil!
       (move-to-column (or csv--header-hscroll 0))
-      (let ((str (buffer-substring (point) (line-end-position)))
+      (let ((str (replace-regexp-in-string
+		  "%" "%%" (buffer-substring (point) (line-end-position))))
             (i 0))
         (while (and i (< i (length str)))
           (let ((prop (get-text-property i 'display str)))
